#![feature(iter_intersperse)]
#![feature(trait_alias)]
use std::{fs, io::{BufRead, Write}, path::Path, process::{self, Stdio}};

fn main() {

    let stdin = std::io::stdin();
    let mut in_lock = stdin.lock();

    let stdout = std::io::stdout();
    let mut out_lock = stdout.lock();

    let commands = &Command::commands();
    println!("Loaded {} commands", commands.len());

    loop {

        let mut buf = String::new();

        let ps1 = "> ".as_bytes();
        out_lock.write(ps1).unwrap();
        out_lock.flush().unwrap();

        in_lock.read_line(&mut buf).expect("Failed to read in to buffer.");
        let trimmed_input = buf.trim();

        let o_choice = Command::choose(trimmed_input, commands);
        if let Some(choice) = o_choice {
            let words = trimmed_input.split(' ').collect::<Vec<&str>>();
            choice.call(words);
        }
    }
}

type Param<'a> = Vec<&'a str>;
trait Run = Fn(Param);
type Cmd = Box<dyn Run>;

// #[derive(Debug)]
struct Command {
    name: String,
    _sub_commands: Vec<Command>,
    cmd: Cmd,
}

impl Command {
    fn init<T>(name: T, sub_commands: Vec<Self>, f: Cmd) -> Self where T: ToString {
        Self {
            name: name.to_string(),
            _sub_commands: sub_commands,
            cmd: f,
        }
    }

    fn call(&self, words: Param) {
        let x = &self.cmd;
        x(words)
    }

    fn choose<'a>(text: &str, choices: &'a Vec<Self>) -> Option<&'a Self> {

        // TODO there seems to be duplicates in here?

        // See what command the user is trying to call
        // iter thru all the commands
        let mut ranked = choices.iter().fold(Vec::new(), |mut v, cmd| {
            // count how many chars match from the command to the word
            let matches = text.chars().zip(cmd.name.chars()).fold(0, |mut f, (word, cmd)| {
                if word == cmd { f += 1; }
                f
            });
            v.push((cmd, matches));
            v
        });

        // sort by most matches
        ranked.sort_by(|a,b| b.1.cmp(&a.1));

        if ranked.len() > 1 {
            // If the top two command choices are ranked the same, we don't know which to choose!
            let top = ranked[0];
            let second = ranked[0];
            if top.1 == second.1 {
                println!("Ambigous command '{}'.\nCould be any of {}, {}, etc...", text, top.0.name, second.0.name);
                return None;
            }
        }

        if let Some((cmd, _)) = ranked.first() {
            // let words = text.split(' ').collect::<Vec<&str>>();

            return Some(cmd);
        }
        None
    }

    fn commands() -> Vec<Self> {
        let path = Path::new("/bin");
        let res = fs::read_dir(path).expect("Failed to read /bin");
    
        // load all the programs from /bin
        let bin = res.fold(Vec::new(), |mut v, contents| {
            if let Ok(cmd) = contents {
                if let Ok(t) = cmd.file_type() {
                    if t.is_file() {
                        // v.push(cmd.path());
                        let cmd = Command::init(
                            cmd.file_name().into_string().unwrap(),
                            vec![],
                            Box::new(move |args| {
                                let name = cmd.path().to_str().unwrap().to_owned();
                                let out = process::Command::new(name)
                                    .args(args)
                                    // FIXME not printing to out
                                    .stdout(Stdio::inherit())
                                    .output()
                                    .expect("Failed to call external program");
                                let y = out.stdout.iter().map(|b| *b as char).collect::<String>();
                                // TODO load into env
                                // let status = out.status.code();
                                println!("{y}"); 
                                }
                                )
                            );
                        v.push(cmd);
                        }
                    };
                } ;
            v
        });
    
        // load custom programs
        let mut commands = vec![
            Command::init("exit", vec![], Box::new(|_| {
                std::process::exit(0);
            })),
            Command::init("custom_echo", vec![], Box::new(|f| {
                let f:Vec<&str>=f;
                let msg = &f[1..].iter().map(|s| s.to_owned()).intersperse(" ").collect::<String>();
                println!("{msg}");
            })),
        ];

        // combine with a move
        commands.extend(bin);
        commands
    }
}
    
